module Main where

import Data.List
import TRS
import Parser
import qualified PRLCert
import qualified PCPSCert
import ComposedCriteria
import qualified SN
import qualified NonConfluenceCert
import CeTA
import System.Environment

-- version
version :: String
version = "0.10"

-- main
help :: IO ()
help = do
  putStrLn "hakusan [option]* [file.trs]"
  putStrLn ("version " ++ version)
  putStrLn ""
  putStrLn "Options:"
  putStrLn "  -tt <s>   uses termination tool <s> with WST format"
  putStrLn "  -ttx <s>  uses termination tool <s> with XML format"
  putStrLn "  -smt <s>  uses SMT solver <s>"
  putStrLn "  -R <k>    rule labeling (Zankl et al. 2015; Shintani and Hirokawa 2022)"
  putStrLn "  -C <k>    successive application of parallel critical pair systems (Shintani and Hirokawa 2022)"
  putStrLn "  -RC <k>   rule labeling and then parallell critical pair systems"
  putStrLn "  -CR <k>   critical pair systems and then rule labeling"
  putStrLn "  -full     combination of R and C"
  putStrLn "  -v        version"
  putStrLn "  -h        help"

data Config = Config {
  _criterion :: CriterionCert,
  _smt :: String,
  _termination :: SN.Tool
}

showResult :: TRS -> Proof -> String
showResult trsR proof@(CRProof _)    = "YES\n" ++ CeTA.showCertificate (trsR, proof)
showResult trsR proof@(CRDisproof _) = "NO\n"  ++ CeTA.showCertificate (trsR, proof)

prove :: CriterionCert -> String -> IO ()
prove cr file = do 
  spec <- readSpecFile file
  case spec of
    Left e     -> print e
    Right trsR -> do
      m1 <- NonConfluenceCert.disprove trsR
      case m1 of
        Just proof -> putStr (showResult trsR proof)
        Nothing -> do
          m2 <- cr trsR
          case m2 of
            Just proof -> putStr (showResult trsR proof)
            Nothing    -> putStrLn "MAYBE"

parseArgs :: Config -> [String] -> IO ()
parseArgs c ("-smt"   : s : args) =
  parseArgs (c { _smt = s }) args
parseArgs c ("-tt"    : s : args) =
  parseArgs (c { _termination = (SN.WST, s) }) args
parseArgs c ("-ttx"   : s : args) =
  parseArgs (c { _termination = (SN.XML, s) }) args
parseArgs c ("-R" : sk : args) = parseArgs (c {_criterion = cr}) args
  where 
    cr trsR = PRLCert.prl (_smt c) ks trsR
    k = read sk :: Int
    ks = if k < 2 then [k] else [2,k]
parseArgs c ("-C" : sk : args) = parseArgs (c {_criterion = cr}) args
  where cr trsR = PCPSCert.pcpsPlus (SN.snCert (_termination c)) (read sk) trsR
parseArgs c ("-RC" : sk : args) = parseArgs (c {_criterion = cr}) args
  where cr trsR = prlAndPcps (_smt c) (_termination c) (read sk) trsR
parseArgs c ("-CR" : sk : args) = parseArgs (c {_criterion = cr}) args
  where cr trsR = pcpsAndPrl (_smt c) (_termination c) (read sk) trsR
parseArgs c ("-full" : args) = parseArgs (c {_criterion = cr}) args
  where cr trsR = full (_smt c) (_termination c) 5 trsR
parseArgs c ["-h"] = help
parseArgs c ["-v"] = putStrLn version
parseArgs c [file] = prove (_criterion c) file
parseArgs _ _ = help

defaultConfig :: Config
defaultConfig = Config {
  _smt         = smt,
  _termination = termination,
  _criterion   = full smt termination 5
}
  where 
    smt = "z3"
    termination = (SN.WST, "ttt2c.sh")

main :: IO ()
main = do
  args <- getArgs
  parseArgs defaultConfig args
