module ReductionMethod where

import Data.List
import Term
import Rule
import TRS
import Rewriting
import PCP
import PRLCert
import SMT
import CeTA

type JoinSequence = (Sequence, Sequence)


termsInSequence :: Sequence -> [Term]
termsInSequence (steps, _) = [ s | (_, s) <- steps ]

intermediates :: [Step] -> [Step] -> [Term]
intermediates []     stepsL = [ t | (_, t) <- stepsL ]
intermediates stepsR stepsL =
  [ t | (_, t) <- reverse (tail stepsR) ++ stepsL ]

joinSequences' :: Int -> TRS -> Term -> Term -> [(TRS, [Term])]
joinSequences' n trsR s t
  | s == t    = [ ([], []) ]
  | otherwise =  
      [ (rules, ReductionMethod.intermediates stepsR stepsL)
      | (stepsR, u) <- reverseSequences n trsR s,
        (stepsL, v) <- reverseSequences n trsR t,
        u == v,
        let rules = nub [ rule | (rule, _) <- stepsR ++ stepsL ] ]

smallest' :: Int -> ([a], b) -> [([a], b)] -> ([a], b)
smallest' _ pair [] = pair
smallest' n pair ((xs, y) : list)
  | n > m     = smallest' m (xs, y) list
  | otherwise = smallest' n pair list
  where
    m = length xs

smallest :: [([a], b)] -> ([a], b)
smallest [] = error "smallest"
smallest ((xs, y) : list) = smallest' (length xs) (xs, y) list

joinSequences :: [Int] -> TRS -> Term -> Term -> Maybe (TRS, [Term])
joinSequences []       _    _ _ = Nothing
joinSequences (k : ks) trsR s t = 
  case joinSequences' k trsR s t of
    []   -> joinSequences ks trsR s t
    list -> Just (smallest list)

closingSystem' :: [Int] -> TRS -> [Peak] -> Maybe (TRS, [Diagram])
closingSystem' _ _    [] = Just ([], [])
closingSystem' ks trsR (peak@(t, _, _, _, _, _, u) : peaks) = do
  (trsC1, vs) <- joinSequences ks trsR t u
  (trsC2, joinsRS) <- closingSystem' ks trsR peaks
  return (union trsC1 trsC2, (peak, vs) : joinsRS)

closingSystem :: [Int] -> TRS -> Maybe (TRS, [Diagram])
closingSystem ks trsR = closingSystem' ks trsR (pcpeak trsR)

-- rule in C
inC :: Labeling Formula -> Rule -> Formula
inC phi rule
  | Just f <- lookupBy Rule.variant rule phi = f
  | otherwise                                = error "ruleInC"

-- f in Fun(C)
inFunC :: String -> Formula
inFunC f = FVar ("y_" ++ f)

-- C is a superset of C0, 
-- C /= R,
-- for every l -> r in R
--   l ->* C r or 
--   not (Fun(l) is a subset of Fun(C0)) 
constraint :: Labeling Formula -> Int -> TRS -> TRS -> Formula
constraint phi k trsR trsC0 =
  conj [conj [ inC phi rule | rule <- trsC0 ],              
        disj [ neg (inC phi rule) | rule <- trsR ], 
        -- if a rule in C then all variables related symbols in rule should be true
        conj [disj [neg (inC phi rule), 
                    conj [ inFunC f | f <- Rule.functions rule ]]
             | rule <- trsR ],
        -- l ->* C r or not (Fun(l) is a subset of Fun(C0)) for every l -> r in R
        conj [ disj [disj [ conj [ inC phi rule | (rule, _) <- steps ]
                          | (steps, u) <- reverseSequences k trsR l, r == u ],
                     neg (conj [ inFunC f | f <- Term.functions l])]
             | (l, r) <- trsR,
               notElem (l, r) trsC0 ] ]

decode :: Model -> Labeling Formula -> TRS
decode model phi =
  [ rule | (rule, f) <- phi, evalFormula model f ]
  
extend :: String -> [Int] -> TRS -> TRS -> IO (Maybe TRS)
extend _   []       _    _     = return Nothing
extend smt (k : ks) trsR trsC0 = do
  m <- sat smt (ReductionMethod.constraint phi k trsR trsC0)
  case m of
    Nothing    -> extend smt ks trsR trsC0
    Just model -> return (Just (decode model phi))
  where
    phi = [ (rule, FVar ("x_" ++ show i)) | (rule, i) <- zip trsR [0..] ]


reduce :: String -> [Int] -> CriterionCert -> CriterionCert
reduce smt ks cr trsR
  | TRS.leftLinear trsR, 
    Just (trsC0, diagrams) <- closingSystem ks trsR = do
      m1 <- extend smt ks trsR trsC0
      case m1 of
        Nothing -> cr trsR
        Just trsC -> do
          m2 <- reduce smt ks cr trsC
          case m2 of
            Nothing -> cr trsR
            Just (CRDisproof crProofForC) -> return Nothing
            Just (CRProof crProofForC) ->
              let crProof = ReductionMethod {
                _trsR = trsR,
                _trsC = trsC,
                _diagrams = diagrams,
                _crProofForC = crProofForC
              }
              in
                return (Just (CRProof crProof))
  | otherwise = cr trsR
